@@ -1,4 +1,4 @@
-# Copyright 2019 The Keras Tuner Authors
+# Copyright 2019 The KerasTuner Authors
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -18,10 +18,10 @@ from concurrent import futures
 
 import grpc
 
-from ..engine import hyperparameters as hp_module
-from ..engine import trial as trial_module
-from ..protos import service_pb2
-from ..protos import service_pb2_grpc
+from keras_tuner.engine import hyperparameters as hp_module
+from keras_tuner.engine import trial as trial_module
+from keras_tuner.protos import service_pb2
+from keras_tuner.protos import service_pb2_grpc
 
 
 class OracleServicer(service_pb2_grpc.OracleServicer):
