@@ -1,4 +1,4 @@
-# Copyright 2019 The Keras Tuner Authors
+# Copyright 2019 The KerasTuner Authors
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -11,21 +11,25 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-"Keras Tuner class."
+"The Tuner class."
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
 
+import contextlib
 import copy
+import gc
 import os
+import traceback
 
+import numpy as np
 import tensorflow as tf
 from tensorboard.plugins.hparams import api as hparams_api
+from tensorflow import keras
 
-from . import base_tuner
-from . import hypermodel as hm_module
-from . import tuner_utils
+from keras_tuner import config as config_module
+from keras_tuner.engine import base_tuner
+from keras_tuner.engine import tuner_utils
+
+MAX_FAIL_STREAK = 5
 
 
 class Tuner(base_tuner.BaseTuner):
@@ -36,9 +40,11 @@ class Tuner(base_tuner.BaseTuner):
     tuners can be created by subclassing the class.
 
     Args:
-        oracle: Instance of Oracle class.
-        hypermodel: Instance of HyperModel class (or callable that takes
-            hyperparameters and returns a Model instance).
+        oracle: Instance of `Oracle` class.
+        hypermodel: Instance of `HyperModel` class (or callable that takes
+            hyperparameters and returns a `Model` instance). It is optional
+            when `Tuner.run_trial()` is overriden and does not use
+            `self.hypermodel`.
         max_model_size: Integer, maximum number of scalars in the parameters of
             a model. Models larger than this are rejected.
         optimizer: Optional `Optimizer` instance.  May be used to override the
@@ -66,6 +72,12 @@ class Tuner(base_tuner.BaseTuner):
         overwrite: Boolean, defaults to `False`. If `False`, reloads an
             existing project of the same name if one is found. Otherwise,
             overwrites the project.
+        executions_per_trial: Integer, the number of executions (training a
+            model from scratch, starting from a new initialization) to run per
+            trial (model configuration). Model metrics may vary greatly
+            depending on random initialization, hence it is often a good idea
+            to run several executions per trial in order to evaluate the
+            performance of a given set of hyperparameter values.
 
     Attributes:
         remaining_trials: Number of trials remaining, `None` if `max_trials` is
@@ -75,7 +87,7 @@ class Tuner(base_tuner.BaseTuner):
     def __init__(
         self,
         oracle,
-        hypermodel,
+        hypermodel=None,
         max_model_size=None,
         optimizer=None,
         loss=None,
@@ -86,17 +98,14 @@ class Tuner(base_tuner.BaseTuner):
         logger=None,
         tuner_id=None,
         overwrite=False,
+        executions_per_trial=1,
     ):
-
-        # Subclasses of `KerasHyperModel` are not automatically wrapped.
-        if not isinstance(hypermodel, hm_module.KerasHyperModel):
-            hypermodel = hm_module.KerasHyperModel(
-                hypermodel,
-                max_model_size=max_model_size,
-                optimizer=optimizer,
-                loss=loss,
-                metrics=metrics,
-                distribution_strategy=distribution_strategy,
+        if hypermodel is None and self.__class__.run_trial is Tuner.run_trial:
+            raise ValueError(
+                "Received `hypermodel=None`. We only allow not specifying "
+                "`hypermodel` if the user defines the search space in "
+                "`Tuner.run_trial()` by subclassing a `Tuner` class without "
+                "using a `HyperModel` instance."
             )
 
         super(Tuner, self).__init__(
@@ -108,6 +117,17 @@ class Tuner(base_tuner.BaseTuner):
             overwrite=overwrite,
         )
 
+        if isinstance(oracle.objective, list) and len(oracle.objective) > 1:
+            raise ValueError(
+                "Multi-objective is not supported, found: {}".format(
+                    oracle.objective
+                )
+            )
+
+        self.max_model_size = max_model_size
+        self.optimizer = optimizer
+        self.loss = loss
+        self.metrics = metrics
         self.distribution_strategy = distribution_strategy
 
         # Support multi-worker distribution strategies w/ distributed tuning.
@@ -127,7 +147,71 @@ class Tuner(base_tuner.BaseTuner):
 
         self.tuner_id = tuner_id or self.tuner_id
 
-    def _build_and_fit_model(self, trial, fit_args, fit_kwargs):
+        self.executions_per_trial = executions_per_trial
+        # This is the `step` that will be reported to the Oracle at the end
+        # of the Trial. Since intermediate results are not used, this is set
+        # to 0.
+        self._reported_step = 0
+
+    def _build_hypermodel(self, hp):
+        with maybe_distribute(self.distribution_strategy):
+            model = self.hypermodel.build(hp)
+            self._override_compile_args(model)
+            return model
+
+    def _try_build(self, hp):
+        for i in range(MAX_FAIL_STREAK + 1):
+            # clean-up TF graph from previously stored (defunct) graph
+            keras.backend.clear_session()
+            gc.collect()
+
+            # Build a model, allowing max_fail_streak failed attempts.
+            try:
+                model = self._build_hypermodel(hp)
+            except:
+                if config_module.DEBUG:
+                    traceback.print_exc()
+
+                print("Invalid model %s/%s" % (i, MAX_FAIL_STREAK))
+
+                if i == MAX_FAIL_STREAK:
+                    raise RuntimeError("Too many failed attempts to build model.")
+                continue
+
+            # Stop if `build()` does not return a valid model.
+            if not isinstance(model, keras.models.Model):
+                raise RuntimeError(
+                    "Model-building function did not return "
+                    "a valid Keras Model instance, found {}".format(model)
+                )
+
+            # Check model size.
+            size = maybe_compute_model_size(model)
+            if self.max_model_size and size > self.max_model_size:
+                print("Oversized model: {} parameters -- skipping".format(size))
+                if i == MAX_FAIL_STREAK:
+                    raise RuntimeError("Too many consecutive oversized models.")
+                continue
+            break
+        return model
+
+    def _override_compile_args(self, model):
+        with maybe_distribute(self.distribution_strategy):
+            if self.optimizer or self.loss or self.metrics:
+                compile_kwargs = {
+                    "optimizer": model.optimizer,
+                    "loss": model.loss,
+                    "metrics": model.metrics,
+                }
+                if self.loss:
+                    compile_kwargs["loss"] = self.loss
+                if self.optimizer:
+                    compile_kwargs["optimizer"] = self.optimizer
+                if self.metrics:
+                    compile_kwargs["metrics"] = self.metrics
+                model.compile(**compile_kwargs)
+
+    def _build_and_fit_model(self, trial, *args, **kwargs):
         """For AutoKeras to override.
 
         DO NOT REMOVE this function. AutoKeras overrides the function to tune
@@ -139,73 +223,109 @@ class Tuner(base_tuner.BaseTuner):
             trial: A `Trial` instance that contains the information needed to
                 run this trial. `Hyperparameters` can be accessed via
                 `trial.hyperparameters`.
-            fit_args: Positional arguments passed by `search`.
-            fit_kwargs: Keyword arguments passed by `search`.
+            *args: Positional arguments passed by `search`.
+            **kwargs: Keyword arguments passed by `search`.
 
         Returns:
             The fit history.
         """
-        model = self.hypermodel.build(trial.hyperparameters)
-        return model.fit(*fit_args, **fit_kwargs)
+        hp = trial.hyperparameters
+        model = self._try_build(hp)
+        return self.hypermodel.fit(hp, model, *args, **kwargs)
 
-    def run_trial(self, trial, *fit_args, **fit_kwargs):
+    def run_trial(self, trial, *args, **kwargs):
         """Evaluates a set of hyperparameter values.
 
         This method is called multiple times during `search` to build and
-        evaluate the models with different hyperparameters.
+        evaluate the models with different hyperparameters and return the
+        objective value.
 
-        The method is responsible for reporting metrics related to the `Trial`
-        to the `Oracle` via `self.oracle.update_trial`.
+        Example:
+
+        You can use it with `self.hypermodel` to build and fit the model.
+
+        ```python
+        def run_trial(self, trial, *args, **kwargs):
+            hp = trial.hyperparameters
+            model = self.hypermodel.build(hp)
+            return self.hypermodel.fit(hp, model, *args, **kwargs)
+        ```
+
+        You can also use it as a black-box optimizer for anything.
+
+        ```python
+        def run_trial(self, trial, *args, **kwargs):
+            hp = trial.hyperparameters
+            x = hp.Float("x", -2.0, 2.0)
+            y = x * x + 2 * x + 1
+            return y
+        ```
 
         Args:
             trial: A `Trial` instance that contains the information needed to
-                run this trial. `Hyperparameters` can be accessed via
+                run this trial. Hyperparameters can be accessed via
                 `trial.hyperparameters`.
-            *fit_args: Positional arguments passed by `search`.
-            **fit_kwargs: Keyword arguments passed by `search`.
+            *args: Positional arguments passed by `search`.
+            **kwargs: Keyword arguments passed by `search`.
+
+        Returns:
+            A `History` object, which is the return value of `model.fit()`, a
+            dictionary, a float, or a list of one of these types.
+
+            If return a dictionary, it should be a dictionary of the metrics to
+            track. The keys are the metric names, which contains the
+            `objective` name. The values should be the metric values.
+
+            If return a float, it should be the `objective` value.
+
+            If evaluating the model for multiple times, you may return a list
+            of results of any of the types above. The final objective value is
+            the average of the results in the list.
         """
-        # Handle any callbacks passed to `fit`.
-        copied_fit_kwargs = copy.copy(fit_kwargs)
-        callbacks = fit_kwargs.pop("callbacks", [])
-        callbacks = self._deepcopy_callbacks(callbacks)
-        self._configure_tensorboard_dir(callbacks, trial)
-        # `TunerCallback` calls:
-        # - `Tuner.on_epoch_begin`
-        # - `Tuner.on_batch_begin`
-        # - `Tuner.on_batch_end`
-        # - `Tuner.on_epoch_end`
-        # These methods report results to the `Oracle` and save the trained Model. If
-        # you are subclassing `Tuner` to write a custom training loop, you should
-        # make calls to these methods within `run_trial`.
-        callbacks.append(tuner_utils.TunerCallback(self, trial))
-        copied_fit_kwargs["callbacks"] = callbacks
+        model_checkpoint = keras.callbacks.ModelCheckpoint(
+            filepath=self._get_checkpoint_fname(trial.trial_id, self._reported_step),
+            monitor=self.oracle.objective.name,
+            mode=self.oracle.objective.direction,
+            save_best_only=True,
+            save_weights_only=True,
+        )
+        original_callbacks = kwargs.pop("callbacks", [])
 
-        self._build_and_fit_model(trial, fit_args, copied_fit_kwargs)
+        # Run the training process multiple times.
+        histories = []
+        for execution in range(self.executions_per_trial):
+            copied_kwargs = copy.copy(kwargs)
+            callbacks = self._deepcopy_callbacks(original_callbacks)
+            self._configure_tensorboard_dir(callbacks, trial, execution)
+            callbacks.append(tuner_utils.TunerCallback(self, trial))
+            # Only checkpoint the best epoch across all executions.
+            callbacks.append(model_checkpoint)
+            copied_kwargs["callbacks"] = callbacks
+            obj_value = self._build_and_fit_model(trial, *args, **copied_kwargs)
 
-    def save_model(self, trial_id, model, step=0):
-        epoch = step
-        self._checkpoint_model(model, trial_id, epoch)
-        # TODO: save the top epoch checkpoints instead of last ones.
-        epoch_to_delete = epoch - self._save_n_checkpoints
-        best_epoch = 0
-        if epoch > 0:
-            # TODO: `get_best_models` would load the `best_step` checkpoint after
-            # training. It would break if oracle picks a different `best_step` than
-            # `metrics.get_best_step` since it might be deleted due to it was
-            # not the `best_epoch` during the training.
-            best_epoch = self.oracle.get_trial(trial_id).metrics.get_best_step(
-                self.oracle.objective.name
-            )
-        if epoch > self._save_n_checkpoints and epoch_to_delete != best_epoch:
-            self._delete_checkpoint(trial_id, epoch_to_delete)
+            # objective left unspecified,
+            # and objective value is not a single float.
+            if (
+                not isinstance(obj_value, (int, float))
+                and self.oracle.objective.name == "default_objective"
+            ):
+                raise TypeError(
+                    "Expect the return value of `Tuner.run_trial()` or "
+                    "`HyperModel.fit()` to be a single float when "
+                    "`objective` is left unspecified. Recevied return value: "
+                    f"{obj_value}."
+                )
+
+            histories.append(obj_value)
+        return histories
 
     def load_model(self, trial):
-        model = self.hypermodel.build(trial.hyperparameters)
+        model = self._try_build(trial.hyperparameters)
         # Reload best checkpoint. The Oracle scores the Trial and also
         # indicates at what epoch the best value of the objective was
         # obtained.
         best_epoch = trial.best_step
-        with hm_module.maybe_distribute(self.distribution_strategy):
+        with maybe_distribute(self.distribution_strategy):
             model.load_weights(
                 self._get_checkpoint_fname(trial.trial_id, best_epoch)
             )
@@ -245,21 +365,9 @@ class Tuner(base_tuner.BaseTuner):
         pass
 
     def on_epoch_end(self, trial, model, epoch, logs=None):
-        """Called at the end of an epoch.
-
-        Args:
-            trial: A `Trial` instance.
-            model: A Keras `Model`.
-            epoch: The current epoch number.
-            logs: Dict. Metrics for this epoch. This should include
-              the value of the objective for this epoch.
-        """
-        self.save_model(trial.trial_id, model, step=epoch)
-        # Report intermediate metrics to the `Oracle`.
-        status = self.oracle.update_trial(trial.trial_id, metrics=logs, step=epoch)
-        trial.status = status
-        if trial.status == "STOPPED":
-            model.stop_training = True
+        # Intermediate results are not passed to the Oracle, and
+        # checkpointing is handled via a `ModelCheckpoint` callback.
+        pass
 
     def get_best_models(self, num_models=1):
         """Returns the best model(s), as determined by the tuner's objective.
@@ -294,11 +402,13 @@ class Tuner(base_tuner.BaseTuner):
             )
         return callbacks
 
-    def _configure_tensorboard_dir(self, callbacks, trial):
+    def _configure_tensorboard_dir(self, callbacks, trial, execution=0):
         for callback in callbacks:
             if callback.__class__.__name__ == "TensorBoard":
                 # Patch TensorBoard log_dir and add HParams KerasCallback
-                logdir = self._get_tensorboard_dir(callback.log_dir, trial.trial_id)
+                logdir = self._get_tensorboard_dir(
+                    callback.log_dir, trial.trial_id, execution
+                )
                 callback.log_dir = logdir
                 hparams = tuner_utils.convert_hyperparams_to_hparams(
                     trial.hyperparameters
@@ -309,8 +419,8 @@ class Tuner(base_tuner.BaseTuner):
                     )
                 )
 
-    def _get_tensorboard_dir(self, logdir, trial_id):
-        return os.path.join(str(logdir), str(trial_id))
+    def _get_tensorboard_dir(self, logdir, trial_id, execution):
+        return os.path.join(str(logdir), str(trial_id), "execution" + str(execution))
 
     def _get_checkpoint_dir(self, trial_id, epoch):
         return os.path.join(
@@ -332,3 +442,21 @@ class Tuner(base_tuner.BaseTuner):
 
     def _delete_checkpoint(self, trial_id, epoch):
         tf.io.gfile.rmtree(self._get_checkpoint_dir(trial_id, epoch))
+
+
+def maybe_compute_model_size(model):
+    """Compute the size of a given model, if it has been built."""
+    if model.built:
+        params = [keras.backend.count_params(p) for p in model.trainable_weights]
+        return int(np.sum(params))
+    return 0
+
+
+@contextlib.contextmanager
+def maybe_distribute(distribution_strategy):
+    """Distributes if distribution_strategy is set."""
+    if distribution_strategy is None:
+        yield
+    else:
+        with distribution_strategy.scope():
+            yield
